import json
import re
import sys
import socket

from base64 import b64decode

sys.path.insert(0, '/motadata/motadata/motadata python engine/')

from com.motadata.util.sdkconstants import TraceOrgSDKConstant
from com.motadata.commons.sshclient import TraceOrgSSHClient
from com.motadata.util.loggerutil import TraceOrgLogger

_logger = TraceOrgLogger("Linux Process Metric Collection")

def collect(context):

    result = {}

    ssh_client = None

    process = ""

    host = ""

    if context and context.get('plugin-context'):

        plugin_context = context.get("plugin-context")

        try:

            timeout = 60

            if context.get("metric-collect-timeout"):

                timeout = int(context.get("metric-collect-timeout"))

            plugin_context['timeout'] = timeout

            process = str(plugin_context.get("process")).strip()

            processCommand = str(plugin_context.get("process-command")).strip()

            host = str(plugin_context.get('host')).strip()

            ssh_client = get_ssh_client(plugin_context)

            if ssh_client is not None and ssh_client.init():

                if process is not None and len(process) > 0 and processCommand is not None and len(processCommand) > 0:

                    output = ssh_client.execute_command("cat /proc/cpuinfo | grep processor | wc -l")

                    if output is not None and len(str(output).strip()) > 0 :

                        cpuCores = int(output.strip())

                        setProcessDetails(ssh_client, result, process, cpuCores, processCommand, host)

                    else :

                        result[TraceOrgSDKConstant.ERROR_CODE] = "No response found for the process {} of host {}.".format(process, host)
            else :

                result[TraceOrgSDKConstant.ERROR_CODE] = "SSH client is not initialised for the host : " + host

        except Exception:

            _logger.debug("Failed to collect linux process metric for the process {} of host {} .".format(process, host))

            _logger.error()

            result[TraceOrgSDKConstant.ERROR_CODE] = "Failed to collect linux process metric for the process {} of host {} .".format(process, host)

        finally:

            if ssh_client:

                ssh_client.destroy()

    return result

def get_ssh_client(plugin_context):

    ssh_client = None

    try:

        user_name = str(plugin_context.get('username')).strip()

        password = str(plugin_context.get('password')).strip()

        port = int(plugin_context.get('port'))

        timeout = int(plugin_context.get('timeout'))

        host = str(plugin_context.get('host')).strip()

        private_key = str(plugin_context.get('private-key-name')).strip()

        passphrase = str(plugin_context.get('pass-phrase')).strip()

        if user_name is None or password is None or port is None or timeout is None or host is None:

            _logger.debug("Required parameter is missing while creating SSH client")

        else:

            if private_key and len(private_key) > 0:

                ssh_client = TraceOrgSSHClient(host=host,port=port,username=user_name,password=password,
                                               timeout=timeout,key_file=private_key,passphrase=passphrase)

            else:

                ssh_client = TraceOrgSSHClient(host=host, port=port,username=user_name,
                                               password=password,timeout=timeout)

    except Exception:

        _logger.error()

    return ssh_client

def setProcessDetails(ssh_client, result, process, cpuCores, processCommand, host) :

    try:

        processPattern = re.compile("(\\w.*)\\s+(\\d+)\\s+(\\d+.\\d+)\\s+(\\d+.\\d+)\\s+(.*)")

        diskPattern = re.compile("(\\d+)\\s+(\\d+)(?:.\\d+)\\s+(\\d+)(?:.\\d+)")

        processMemory = 0

        processExistoutput = ssh_client.execute_command("pgrep '" + process + "' | awk '{print $1}'")

        if processExistoutput is not None and len(str(processExistoutput).strip()) > 0 and str(processExistoutput) != '\\n':

            processIdArray = str(processExistoutput).split('\n')

            for processId in processIdArray :

                if processId is not None and str(processId).lower() != "command" and str(processId) != ("\\r") and str(processId) != (""):

                    argumentOutput = ssh_client.execute_command("ps -p " + processId + " -o args")

                    if argumentOutput is not None and argumentOutput != "" and len(str(argumentOutput).strip()) > 0 :

                        processOutputArray = str(argumentOutput).split('\n')

                        for processOutput in processOutputArray :

                            if processOutput is not None and str(processOutput).lower() != "command" and str(processOutput) != ("\\r") and str(processOutput) != (""):

                                if processCommand in str(processOutput):

                                    output = ssh_client.execute_command("ps -up '" + processId + "'| awk '{print $1\" \"$2\" \"$3\" \"$4\" \"$12}'")

                                    for processMatcher in re.finditer(processPattern, output):

                                        result["Process"] = process

                                        if ("motadata-") in processOutput :

                                            result["Process Command Line"] = processOutput.split(" ")[0]

                                        else :

                                            result["Process Command Line"] = processOutput

                                        result["Process User"] = str(processMatcher.group(1))

                                        result["Process Id"] = int(processMatcher.group(2))

                                        threadCount = ssh_client.execute_command("ps hH p " + processMatcher.group(2)+ " | wc -l")

                                        if threadCount is not None and str(threadCount) != "" and len(str(threadCount).strip()) > 0:

                                            result["Process Threads"] = int(str(threadCount).strip())

                                        processCPU = float(processMatcher.group(3).strip())

                                        result["Process CPU (%)"] = int(processCPU/cpuCores)

                                        result["Process CPU Time (%)"] = int(processCPU)

                                        memoryOutput = ssh_client.execute_command("free -m |grep -i mem |awk '{print $2}'")

                                        if memoryOutput is not None and str(memoryOutput).strip() != ("") and len(str(memoryOutput).strip()) > 0 :

                                            processMemory = int(str(memoryOutput).strip()) * 1024

                                        if processMemory != -1 :

                                            result["Process Memory (%)"] = int(float(processMatcher.group(4)))

                                            processMemoryKB = (float(processMatcher.group(4)) * processMemory) / 100

                                            result["Process Memory (KB)"] = int(processMemoryKB)

                                        diskOutput = ssh_client.execute_command("pidstat -p " + processMatcher.group(2) + " -d | awk '{print $4\" \"$5\" \"$6}'")

                                        if diskOutput is not None and str(diskOutput).strip() != "" and len(str(diskOutput).strip()) > 0 :

                                            diskOutPutArray = str(diskOutput).split('\n')

                                            for diskSingleElement in diskOutPutArray :

                                                for processDiskMatcher in re.finditer(diskPattern, diskSingleElement):

                                                    if ( str(result.get('Process Id')) == processDiskMatcher.group(1).strip()) :

                                                        processDiskReads = int(float(processDiskMatcher.group(2).strip()))

                                                        processDiskWrites = int(float(processDiskMatcher.group(3).strip()))

                                                        result["Process I/O (Bytes/sec)"] = (processDiskReads + processDiskWrites) * 1024

                if result is not None and len(result) > 0 :

                    break

    except Exception:

        _logger.debug("Error occur while setting field in the linux process metrix for process {} of host {}".format(process, host))

        result[TraceOrgSDKConstant.ERROR_CODE] = "Exception thrown for the process {} of host {}".format(process, host)

        _logger.error()

    finally:

        if len(result) <= 0 :

            _logger.debug("SSH command return blank result while setting field in the linux process metrix for process {} of host {}".format(process, host))

            result[TraceOrgSDKConstant.ERROR_CODE] = "No response found for the process {} of host {}.".format(process, host)

def discover(ssh_client, plugin_context):

    result = {}

    try:

        process = str(plugin_context.get("process")).strip()

        processCommand = str(plugin_context.get("process-command")).strip()

        host = str(plugin_context.get("host")).strip()

        if process is not None and len(process) > 0 and processCommand is not None and len(processCommand) > 0:

            if plugin_context.get('monitor_id'):

                checkProcessExist(ssh_client, result, process, processCommand, host)

            else:

                command = "ps -A | grep '" + process + "' | awk '{print $1\" \"$4}'"

                processes = getProcesses(ssh_client, processCommand, command, 1, host)

                setProcessDiscovery(processes, result, host, process)

        elif process is not None and len(process) > 0:

            if plugin_context.get('monitor_id'):

                checkProcessExist(ssh_client, result, process, None, host)

            else:

                command = "ps -A | grep '"+ process +"' | awk '{print $1\" \"$4}'"

                processes = getProcesses(ssh_client, None, command, 0, host)

                setProcessDiscovery(processes, result, host, process)

        elif processCommand is not None and len(processCommand) > 0:

            if plugin_context.get('monitor_id'):

                checkProcessExist(ssh_client, result, None, processCommand, host)

            else:

                command = "top -b -c -n 1 | grep '"+processCommand+"' | awk '{print $1\" \"$12}'"

                processes = getProcesses(ssh_client, processCommand, command, 1, host)

                setProcessDiscovery(processes, result, host, None)

        else:

            output = ssh_client.execute_command("uname")

            if output and len(str(output).strip()) > 0:

                command = "ps -A | awk '{print $1\" \"$4}'"

                processes = getProcesses(ssh_client, None, command, 0, host)

                setProcessDiscovery(processes, result, host, "all")

    except Exception:

        _logger.error()

    return result

def checkProcessExist(ssh_client, result, process, processCommand, host):

    processExists = False

    try:

        processPattern = re.compile("(\\w.*)\\s+(\\d+)\\s+(\\d+.\\d+)\\s+(\\d+.\\d+)\\s+(.*)")

        processExistoutput = ssh_client.execute_command("pgrep '" + process + "' | awk '{print $1}'")

        if processExistoutput is not None and len(str(processExistoutput).strip()) > 0 and str(processExistoutput) != '\\n':

            processIdArray = str(processExistoutput).split('\n')

            for processId in processIdArray :

                if processId is not None and str(processId).lower() != "command" and str(processId) != ("\\r") and str(processId) != (""):

                    argumentOutput = ssh_client.execute_command("ps -p " + processId + " -o args")

                    if argumentOutput is not None and argumentOutput != "" and len(str(argumentOutput).strip()) > 0 :

                        processOutputArray = str(argumentOutput).split('\n')

                        for processOutput in processOutputArray :

                            if processOutput is not None and str(processOutput).lower() != "command" and str(processOutput) != ("\\r") and str(processOutput) != (""):

                                if processCommand in str(processOutput):

                                    output = ssh_client.execute_command("ps -up '" + processId + "'| awk '{print $1\" \"$2\" \"$3\" \"$4\" \"$12}'")

                                    for processMatcher in re.finditer(processPattern, output):

                                        if processMatcher:

                                            processExists = True

                                            break

                if processExists :

                    break

    except Exception:

        _logger.debug("Error occur in 'checkProcessExist' method while discovery for the process {}, host {}.".format(process, host))

        result[TraceOrgSDKConstant.ERROR_CODE] = "Exception thrown while discovery for the process {}, host {}.".format(process, host)

        _logger.error()

    finally:

        if processExists:

            result[TraceOrgSDKConstant.STATUS_CODE] = TraceOrgSDKConstant.STATUS_CODE_REACHABLE

            result[TraceOrgSDKConstant.STATUS] = TraceOrgSDKConstant.SUCCESS

        else:

            result[TraceOrgSDKConstant.STATUS_CODE] = TraceOrgSDKConstant.STATUS_CODE_UNREACHABLE

            result[TraceOrgSDKConstant.STATUS] = TraceOrgSDKConstant.FAILED

            result[TraceOrgSDKConstant.ERROR_CODE] = 'Linux process {} on host {} is not running'.format(process, host)

def getProcesses(ssh_client, processCommand, command, discoveryType, host):

    processes = {}

    try:

        processPattern = re.compile("(\\d+)\\s+(.*)")

        processExistoutput = ssh_client.execute_command(command)

        if processExistoutput is not None and len(str(processExistoutput).strip()) > 0 and str(processExistoutput) != '\\n':

            processArray = str(processExistoutput).split('\n')

            for process in processArray:

                for processMatcher in  re.finditer(processPattern,process):

                    processId = str(processMatcher.group(1)).strip()

                    processName = str(processMatcher.group(2)).strip()

                    if processId is not None and "grep" not in processName.lower() and str(processId) != ("\\r") and str(processId) != ("") and processes.get(processName + "@") is None:

                        argumentOutput = ssh_client.execute_command("ps -p " + processId + " -o args")

                        if argumentOutput is not None and argumentOutput != "" and len(str(argumentOutput).strip()) > 0:

                            processOutputArray = str(argumentOutput).split('\n')

                            for processOutput in processOutputArray:

                                if processOutput is not None and str(processOutput).lower() != "command" and str(processOutput) != ("\\r") and str(processOutput) != (""):

                                    if processes.get(processName):

                                        if discoveryType == 1 and processCommand in str(processOutput):

                                            processes[processName] = processCommand

                                        elif discoveryType == 0:

                                            processes[processName] = str(processOutput).strip()

                                    else :

                                        if discoveryType == 1 and processCommand in str(processOutput):

                                            processes[processName + '@'] = processCommand

                                        elif discoveryType == 0:

                                            processes[processName + '@'] = str(processOutput).strip()

    except Exception:

        _logger.debug("Error occur in 'getProcesses' method while finding process for the host : " + host)

    return processes

def setProcessDiscovery(processes, result, host, process):

    try:

        if processes is not None and len(processes) > 0 :

            result[TraceOrgSDKConstant.STATUS] = TraceOrgSDKConstant.SUCCESS

            resultNodeList = []

            for resultantProcess in processes.keys():

                if "motadata-" in resultantProcess.lower():

                    resultNodeList.append("host : {}, ip : {}, process-command : {}, process : {} ".format(resolve_host(host), resolve_ip(host), resultantProcess.replace("@", ""), resultantProcess.replace("@", "")))

                else:

                    resultNodeList.append("host : {}, ip : {}, process-command : {}, process : {} ".format(resolve_host(host), resolve_ip(host), processes.get(resultantProcess), resultantProcess.replace("@","")))

            result[TraceOrgSDKConstant.RESULT] = json.dumps(resultNodeList)

        elif process is not None and process.lower() == "all":

            result[TraceOrgSDKConstant.STATUS] = TraceOrgSDKConstant.FAILED

            result[TraceOrgSDKConstant.ERROR_CODE] = "Failed to discover all processes for the host :" + host

            _logger.debug("Failed to discover all processes for the host :" + host)

        else:

            result[TraceOrgSDKConstant.STATUS] = TraceOrgSDKConstant.FAILED

            result[TraceOrgSDKConstant.ERROR_CODE] = "Failed to discover process for the host : " + host

            _logger.debug("Failed to discover process for the host :" + host)


    except Exception:

        _logger.error()

    finally:

        if result[TraceOrgSDKConstant.STATUS] is not None and result[TraceOrgSDKConstant.STATUS] == TraceOrgSDKConstant.SUCCESS:

            result[TraceOrgSDKConstant.STATUS_CODE] = TraceOrgSDKConstant.STATUS_CODE_REACHABLE

        else:

            result[TraceOrgSDKConstant.STATUS_CODE] = TraceOrgSDKConstant.STATUS_CODE_UNREACHABLE

def resolve_host(address):

    host = address

    pattern = re.compile("(25[0-5]|2[0-4]\\d|[0-1]?\\d?\\d)(\\.(25[0-5]|2[0-4]\\d|[0-1]?\\d?\\d)){3}")

    try:

        matcher = pattern.match(address)

        if matcher:

            real_host = socket.gethostbyaddr(address)

            host = str(real_host[0]).strip()

    except Exception:

        host = address

    finally:

        return host

def resolve_ip(address):

    ip = address

    pattern = re.compile("(25[0-5]|2[0-4]\\d|[0-1]?\\d?\\d)(\\.(25[0-5]|2[0-4]\\d|[0-1]?\\d?\\d)){3}")

    try:

        matcher = pattern.match(address)

        if matcher is None:

            ip = socket.gethostbyname(address)

    except Exception:

        ip = address

    finally:

        return ip


if __name__ == "__main__":

    result = {}

    ssh_client = None

    if len(sys.argv) > 0 and sys.argv[1] == 'discovery':

        try:

            context = json.loads(b64decode(sys.argv[2]))

            _logger.debug("Pythin Context For Linux Process : " + str(context))

            timeout = 60

            if context.get("metric-collect-timeout"):

                timeout = int(context.get("metric-collect-timeout"))

            context['timeout'] = timeout

            ssh_client = get_ssh_client(context)

            if ssh_client is not None and ssh_client.init():

                result = discover(ssh_client, context)

                if result is not None and len(result) > 0:

                    context[TraceOrgSDKConstant.RESULT] = json.dumps(result)

                else:

                    context[TraceOrgSDKConstant.ERROR_CODE] = "Failed to discover " + context.get('host')

                del context['password']

                del context['username']

            else:

                context[TraceOrgSDKConstant.ERROR_CODE] = "SSH client is not initialised for the host : " + context.get('host')

        except Exception as exception:

            _logger.error()

            context[TraceOrgSDKConstant.ERROR_CODE] = exception.__class__.__name__

        finally:

            if ssh_client is not None:

                ssh_client.destroy()

            print (json.dumps(context))
